########################################################################
#
# File Name: 	        PersistentObject.py
#
# Documentation:	http://docs.ftsuite.com/4ODS/RepositoryObject.py.html
#
"""
Implements the Repository meta-data interface.
WWW: http://4suite.com/4ODS         e-mail: support@4suite.com

Copyright (c) 2000 FourThought Inc, USA.   All Rights Reserved.
See  http://4suite.com/COPYRIGHT  for license and copyright information
"""

from Ft.Ods import Database
from Ft.Ods.StorageManager.Adapters import Constants
from Ft.Ods import IntegrityException
from Ft.Ods.Transaction import TransactionNotInProgress
import tempfile,os
from Ft.Ods.Collections import CollectionBase
class TupleDefinitions:
    TYPE = 1
    READONLY = 2
    ENUM_CLASS = 3
    RELATIONSHIP = 4
    COLLECTION_CLASS = 5





class PersistentObject:
#    _objectIds = {}
#    _literalIds = {}
#    _collectionIds = {}
#    _fileNames = {}

    _persistentType = None
    
    def __init__(self, data=None):
        self.__dict__['_objectIds'] = {}
        self.__dict__['_collectionIds'] = {}
        self.__dict__['_blobIds'] = {}
        self.__dict__['_modifiedAttrs'] = {}
        self.__dict__['_initialized'] = 0
        
        #See if we are in a transaction
        self.__dict__['_db'] = Database.Database()
        if self._db._4ods_inTransaction():
            self.__dict__['_tx'] = self._db.current()
        else:
            self.__dict__['_tx'] = None

        # Initialize member storage
        if data is not None:
            self.__dict__['_initialized'] = 1
            map(lambda typ, name, value, self=self,
                funcs=self._initFuncs:
                funcs[typ](self, name, value),
                self._tupleTypes,
                self._tupleNames,
                data
                )

            if self._tx:
                self._tx._4ods_registerModifiedPersistentObject(self)
        elif self._tx:
            self._tx._4ods_registerNewPersistentObject(self)

        return

    def _pseudo_del(self):
        # Remove all circular references from relationships
        for name,d in self._tupleDefinitions.items():
            if d.get(TupleDefinitions.RELATIONSHIP,0):
                if self.__dict__.has_key(name):
                    obj = self.__dict__[name]
                    del self.__dict__[name]
                    obj._pseudo_del()


    def _4ods_getId(self):
        raise NotImplementedError('Derived class must override')

    def _4ods_isModified(self):
        return len(self.__dict__['_modifiedAttrs']) > 0

    def _4ods_setTransaction(self, newTx, unregister=0):
        if self._tx and unregister:
            self._tx._4ods_unregisterPersistentObject(self)
        self.__dict__['_tx'] = newTx
        if newTx:
            self._tx._4ods_registerPersistentObject(self)

    def _4ods_setBlobId(self, name, bid):
        self.__dict__['_blobIds'][name] = bid

    def delete(self, recurse=0):
        if self._tx:
            self._tx._4ods_deletePersistentObject(self)
        Types = Constants.Types
        lists = Constants.g_listTypes
        for name,data in self._tupleDefinitions.items():
            typ = data[TupleDefinitions.TYPE]
            if data.get(TupleDefinitions.RELATIONSHIP):
                obj = None
                if typ in [Types.OBJECT]:
                    obj = getattr(self,name)
                    if obj:
                        funcName = 'drop_%s' % name
                        func = getattr(self,funcName)
                        func(self.__dict__[name])
                        if recurse:
                            obj.delete(recurse)
                elif lists[typ]:
                    coll = getattr(self,name)
                    while 1:
                        if not len(coll):
                            break
                        element = coll[0]
                        funcName = 'remove_%s' % name
                        func = getattr(self,funcName)
                        func(element)
                        if recurse:
                            element.delete(recurse)
                    coll.delete()
            elif typ == Types.OBJECT:
                obj = getattr(self,name)
                self.__dict__[name] = None
                if self.__dict__['_objectIds'].has_key(name):
                    del self.__dict__['_objectIds'][name]
                if obj:
                    obj.delete(recurse)
            elif typ == Types.STRUCTURE:
                struct = getattr(self,name)
                struct.delete()
            elif typ == Types.BLOB:
                if self.__dict__['_blobIds'].has_key(name):
                    if self.__dict__['_tx']:
                        self.__dict__['_tx']._4ods_deleteBlob(self,name,self.__dict__['_blobIds'][name])
                    del self.__dict__['_blobIds'][name]
            elif lists[typ]:
                if not self.__dict__.has_key(name) and not self.__dict__['_collectionIds'].has_key(name):
                    continue
                obj = getattr(self,name)
                obj.delete(recurse)

    def __getattr__(self, name):
        data = self._tupleDefinitions.get(name)
        if data:
            attr_func = self._getFuncs.get(data[TupleDefinitions.TYPE])
            if attr_func is None:
                raise AttributeError(name)
            value = attr_func(self,name)
            if value is not None:
                self.__dict__[name] = value
            return value
        raise AttributeError(name)


    def __setattr__(self, name, value):
        data = self._tupleDefinitions.get(name)
        if data:
            if self._setFuncs.has_key(data[TupleDefinitions.TYPE]):
                self._setFuncs[data[TupleDefinitions.TYPE]](self,name,value)
                self._modifiedAttrs[name] = 1
                return
            elif data.get(TupleDefinitions.READONLY):
                raise AttributeError('readonly %s' % name)
            self._modifiedAttrs[name] = 1
        self.__dict__[name] = value

    def __delattr__(self,name):
        data = self._tupleDefinitions.get(name)
        if data:
            if data.get(TupleDefinitions.RELATIONSHIP):
                #This means drop ro remove
                #Make sure it is loaded
                obj = getattr(self,name)
                if isinstance(obj,CollectionBase.CollectionBase):
                    func = getattr(self,'remove_%s' % name)
                    for o in self.__dict__[name]:
                        func(o)
                else:
                    o = getattr(self,name)
                    if o:
                        func = getattr(self,'drop_%s' % name)
                        func(o)

            else:
                raise "Cannot delete attributes from a Persistent Object"
        else:
            del self.__dict__[name]

    def _4ods_getFullTuple(self):
        rt = map(lambda typ, name, self=self, funcs=self._getDataFuncs:
                 funcs[typ](self, name),
                 self._tupleTypes,
                 self._tupleNames
                 )

        return rt

    def _4ods_getModifiedTuple(self):
        if not self.__dict__['_modifiedAttrs'].keys():
            # No modifications, we're done!
            return None

        rt = map(lambda typ, name, self=self, funcs=self._getDataFuncs,
                 modified=self.__dict__['_modifiedAttrs']:
                 modified.get(name) and funcs[typ](self, name),
                 self._tupleTypes,
                 self._tupleNames
                 )
        return rt
        
    # Helper for initializing object
    def _4ods_initObject(self, name, value):
        """Initializes an object type for a new object"""
        if value:
            self.__dict__['_objectIds'][name] = value

    # Helper for initializing object
    def _4ods_initEnumeration(self, name, value):
        """Initialize an enumeration type for a new object"""
        if value == -1:
            self.__dict__[name] = None
        elif self._tupleDefinitions[name].has_key(TupleDefinitions.ENUM_CLASS):
            self.__dict__[name]  = self._tupleDefinitions[name][TupleDefinitions.ENUM_CLASS].elements[value]
        else:
            raise "Unknown class based attribute %s" % name
    

    # Helper for initializing object
    def _4ods_initCollection(self, name, value):
        """Initialize a collection type for a new object"""
        if value:
            self.__dict__['_collectionIds'][name] = value

    # Helper for initializing object
    def _4ods_initBlob(self, name, value):
        """Initialize a file type for a new object"""
        self.__dict__['_blobIds'][name] = value

    # Helper for initializing object
    def _4ods_initPrimitive(self, name, value):
        """Initialize a primitive type for a new object"""
        self.__dict__[name] = value

    # Helper for get[Full|Modified]Tuple
    def _4ods_getObjectData(self, name):
        """The OID of the object member or 0 if unaccessed"""
        if self.__dict__.has_key(name) and self.__dict__[name] is not None:
            return self.__dict__[name]._4ods_getId()
        return self.__dict__['_objectIds'].get(name, 0)

    # Helper for get[Full|Modified]Tuple
    def _4ods_getEnumerationData(self, name):
        """The value of enumeration member or -1 if unaccessed"""
        if self.__dict__[name] is None:
            return -1
        return self.__dict__[name]._v
            
    # Helper for get[Full|Modified]Tuple
    def _4ods_getCollectionData(self, name):
        """The CID of the collection member or None if unaccessed or empty"""
        if self.__dict__.has_key(name) and self.__dict__[name].cardinality() > 0:
            return self.__dict__[name]._4ods_getId()
        return self.__dict__['_collectionIds'].get(name,0)

    # Helper for get[Full|Modified]Tuple
    def _4ods_getBlobData(self, name):
        """The contents of the blob member or 0 if unaccessed"""
        return self.__dict__['_blobIds'].get(name, 0)

    # Helper for get[Full|Modified]Tuple
    def _4ods_getPrimitiveData(self, name):
        """The contents of the primitive type member"""
        return self.__dict__[name]

    #Helper Function for getting an object
    def _4ods_getObject(self,name):
        if self.__dict__['_objectIds'].has_key(name):
            if not self._tx:
                raise TransactionNotInProgress()
            if self._persistentType == Constants.ObjectTypes.OBJECT:
                obj = self._tx._4ods_getObject(self.__dict__['_objectIds'][name])
            else:
                obj = self._tx._4ods_getRepositoryObject(self.__dict__['_objectIds'][name])
            self.__dict__[name] = obj
            return obj
        return None

    def _4ods_getCollection(self,name):
        if self.__dict__['_collectionIds'].has_key(name):
            if not self._tx:
                raise TransactionNotInProgress()
            col = self._tx._4ods_getCollection(self.__dict__['_collectionIds'][name])
            self.__dict__[name] = col
            return col

        #Create a new one
        self.__dict__['_modifiedAttrs'][name] = 1
        rt = self._tupleDefinitions[name][TupleDefinitions.COLLECTION_CLASS]()
        self.__dict__[name] = rt
        return rt
                                                        
    def _4ods_getBlob(self,name):
        if self.__dict__['_blobIds'].has_key(name) and self.__dict__['_blobIds'][name]:
            if not self._tx:
                raise TransactionNotInProgress()
            return self._tx._4ods_readBlob(self.__dict__['_blobIds'][name])
        return ''
            


    #Helper Function for setting an object
    def _4ods_setObject(self,name,value):
        #Is it a relationship or an attribute?
        if self._tupleDefinitions[name].get(TupleDefinitions.RELATIONSHIP):
            if self.__dict__.has_key(name) and isinstance(self.__dict__[name],CollectionBase.CollectionBase):
                raise "Cannot use setattr short cut on N-? relationships"
            func = getattr(self,'form_%s' % name)
            func(value)
        else:
            #Set an object attribute
            if value is None:
                self.__dict__[name] = None
                if self.__dict__['_objectIds'].has_key(name):
                    del self.__dict__['_objectIds'][name]
            else:
                self.__dict__['_objectIds'][name] = value._4ods_getId()
            self.__dict__[name] = value

    def _4ods_setBlob(self,name,value):
        if self.__dict__['_tx']:
            if self.__dict__['_blobIds'].has_key(name) and self.__dict__['_blobIds'][name]:
                self.__dict__['_tx']._4ods_registerModifiedBlob(self.__dict__['_blobIds'][name],value)
            else:
                self.__dict__['_tx']._4ods_registerNewBlob(self,name,value)
        self.__dict__[name] = value
            


    _initFuncs = {Constants.Types.OBJECT : _4ods_initObject,
                  Constants.Types.ENUMERATION : _4ods_initEnumeration,
                  Constants.Types.BLOB : _4ods_initBlob,
                  }

    _getDataFuncs = {Constants.Types.OBJECT : _4ods_getObjectData,
                     Constants.Types.ENUMERATION : _4ods_getEnumerationData,
                     Constants.Types.BLOB : _4ods_getBlobData,
                     }

    _getFuncs = {Constants.Types.OBJECT : _4ods_getObject,
                 Constants.Types.BLOB : _4ods_getBlob,
                 }


    _setFuncs = {Constants.Types.OBJECT : _4ods_setObject,
                 Constants.Types.BLOB : _4ods_setBlob,
                 }
    for t,v in Constants.g_listTypes.items():
        if v:
            _initFuncs[t] = _4ods_initCollection
            _getDataFuncs[t] = _4ods_getCollectionData
            _getFuncs[t] = _4ods_getCollection

    for t,v in Constants.g_primitiveTypes.items():
        if v:
            _initFuncs[t] = _4ods_initPrimitive
            _getDataFuncs[t] = _4ods_getPrimitiveData

