########################################################################
#
# File Name:            Postgres.py
#
# Documentation:        http://docs.4suite.com/4RDF/Drivers/Postgres.py.html
#
"""
Interface to Postgres for 4RDF
WWW: http://4suite.com/RDF        e-mail: support@4suite.com

Copyright (c) 2000 FourThought, Inc., USA.  All Rights Reserved.
See  http://4suite.com/COPYRIGHT  for license and copyright information
"""

import sys, string, cPickle, cStringIO
from Ft.Lib.PgDatabase import EscapeQuotes
from Ft.Rdf.Drivers import DataBaseExceptions
from Ft.Rdf import Statement, Model

class Commands:
    LOOKUP_ACL = 0
    NEW_ACL = 1
    ADD_ACL = 2
    CHANGE_ACL = 3
    
    ADD = 10
    REMOVE = 11
    COMPLETE = 12
    CONTAINS = 13
    SIZE = 14
    SIZE_ALL = 15

    BIND = 20
    UNBIND = 21
    LOOKUP = 22
    KEYS = 23
    HAS_KEY = 24


class SqlAdapter:
    def __init__(self, sqlCommands, comparisons):
        self._sql = sqlCommands
        self._cmp = comparisons
        self._db = None
        return

    def begin(self):
        raise NotImplementedError('%s should override this' %
                                  self.__class__.__name__)

    def commit(self):
        raise NotImplementedError('%s should override this' %
                                  self.__class__.__name__)

    def rollback(self):
        raise NotImplementedError('%s should override this' %
                                  self.__class__.__name__)

    def resolveAcl(self, aclNames):
        ids = ['0']
        names = []
        for name in aclNames:
            result = self._sql[Commands.LOOKUP_ACL].query(self._db, name=name)
            if result:
                ids.append(str(result[0][0]))
            else:
                names.append(name)
        return ids, names
            
    def verifyAcl(self, aclName, aclIds, aclNames):
        result = self._sql[Commands.LOOKUP_ACL].query(self._db, name=aclName)
        if result:
            aclId = str(result[0][0])
        else:
            result = self._sql[Commands.NEW_ACL].query(self._db)
            aclId = str(result[0][0])
            self._sql[Commands.ADD_ACL].execute(self._db, id=aclId, name=aclName)

        if aclName in aclNames:
            aclNames.remove(aclName)
            aclIds.append(aclId)

        return aclId

    def add(self, statements, readAcl, deleteAcl, aclIds, aclNames):
        # Takes a list of tuples [(s, p, o, stmtUri, srcUri), ...]
        if not self._db:
            raise DataBaseExceptions.NoTransaction

        readId = readAcl and self.verifyAcl(readAcl, aclIds, aclNames)
        deleteId = deleteAcl and self.verifyAcl(deleteAcl, aclIds, aclNames)

        for s in statements:
            self._sql[Commands.ADD].execute(self._db,
                                            subject=EscapeQuotes(s[0]),
                                            predicate=EscapeQuotes(s[1]),
                                            object=EscapeQuotes(s[2]),
                                            statementUri=EscapeQuotes(s[3]),
                                            sourceUri=EscapeQuotes(s[4]),
                                            readAcl=readId,
                                            deleteAcl=deleteId,
                                            )
        return
        
    def remove(self, statements, aclIds):
        for s in statements:
            self.removePattern(s[0], s[1], s[2], s[3], s[4], {}, aclIds)

    def removePattern(self, subject, predicate, object, statementUri,
                      sourceUri, flags, aclIds):
        if not self._db:
            raise DataBaseException.NoTransaction

        subject = EscapeQuotes(subject)
        predicate = EscapeQuotes(predicate)
        object = EscapeQuotes(object)
        statementUri = EscapeQuotes(statementUri)
        sourceUri = EscapeQuotes(sourceUri)

        # We can use an emtpy string check because of EscapeQuotes()
        command = self._sql[(Commands.REMOVE,
                             subject != '',
                             predicate != '',
                             object != '',
                             statementUri != '',
                             sourceUri != '',
                             )]

        command.execute(self._db,
                        subject=subject,
                        predicate=predicate,
                        object=object,
                        statementUri=statementUri,
                        sourceUri=sourceUri,
                        subjectOp=self._cmp[flags.get('subjectFlags')],
                        predicateOp=self._cmp[flags.get('predicateFlags')],
                        objectOp=self._cmp[flags.get('objectFlags')],
                        statementUriOp=self._cmp[flags.get('statementUriFlags')],
                        sourceUriOp=self._cmp[flags.get('sourceUriFlags')],
                        deleteAcl=string.join(aclIds, ','),
                        )
        return        

    def complete(self, subject, predicate, object, statementUri, sourceUri,
                 flags, aclIds):
        if not self._db:
            raise DataBaseExceptions.NoTransaction

        subject = EscapeQuotes(subject)
        predicate = EscapeQuotes(predicate)
        object = EscapeQuotes(object)
        statementUri = EscapeQuotes(statementUri)
        sourceUri = EscapeQuotes(sourceUri)
        
        # We can use an emtpy string check because of EscapeQuotes()
        command = self._sql[(Commands.COMPLETE,
                             subject != '',
                             predicate != '',
                             object != '',
                             statementUri != '',
                             sourceUri != '',
                             )]
                               
        result = command.query(self._db,
                               subject=subject,
                               predicate=predicate,
                               object=object,
                               statementUri=statementUri,
                               sourceUri=sourceUri,
                               subjectOp=self._cmp[flags.get('subjectFlags')],
                               predicateOp=self._cmp[flags.get('predicateFlags')],
                               objectOp=self._cmp[flags.get('objectFlags')],
                               statementUriOp=self._cmp[flags.get('statementUriFlags')],
                               sourceUriOp=self._cmp[flags.get('sourceUriFlags')],
                               readAcl=string.join(aclIds, ','),
                               )
        return map(lambda x: (x[0], x[1], x[2], x[3], x[4]), result or [])

    def changeAcl(self, subject, predicate, object, statementUri, sourceUri,
                  flags, readAcl, deleteAcl, aclIds, aclNames):
        if not self._db:
            raise DataBaseException.NoTransaction

        subject = EscapeQuotes(subject)
        predicate = EscapeQuotes(predicate)
        object = EscapeQuotes(object)
        statementUri = EscapeQuotes(statementUri)
        sourceUri = EscapeQuotes(sourceUri)

        readId = readAcl and self.verifyAcl(readAcl, aclIds, aclNames)
        deleteId = deleteAcl and self.verifyAcl(deleteAcl, aclIds, aclNames)

        # We can use an emtpy string check because of EscapeQuotes()
        command = self._sql[(Commands.CHANGE_ACL,
                             subject != '',
                             predicate != '',
                             object != '',
                             statementUri != '',
                             sourceUri != '',
                             )]
                               
        command.execute(self._db,
                        subject=subject,
                        predicate=predicate,
                        object=object,
                        statementUri=statementUri,
                        sourceUri=sourceUri,
                        subjectOp=self._cmp[flags.get('subjectFlags')],
                        predicateOp=self._cmp[flags.get('predicateFlags')],
                        objectOp=self._cmp[flags.get('objectFlags')],
                        statementUriOp=self._cmp[flags.get('statementUriFlags')],
                        sourceUriOp=self._cmp[flags.get('sourceUriFlags')],
                        deleteAcl=string.join(aclIds, ','),
                        newReadAcl=readId,
                        newDeleteAcl=deleteId,
                        )
        return        

    def size(self, sourceUri):
        if not self._db:
            raise DataBaseExceptions.NoTransaction

        if sourceUri:
            uri = EscapeQuotes(sourceUri)
            result = self._sql[Commands.SIZE].query(self._db, sourceUri=uri)
        else:
            result = self._sql[Commands.SIZE_ALL].query(self._db)
        return result and result[0][0] or 0

    def contains(self, subject, predicate, object, statementUri, sourceUri, flags):
        if not self._db:
            raise DataBaseExceptions.NoTransaction

        subject = EscapeQuotes(subject)
        predicate = EscapeQuotes(predicate)
        object = EscapeQuotes(object)
        statementUri = EscapeQuotes(statementUri)
        sourceUri = EscapeQuotes(sourceUri)
        
        # We can use an emtpy string check because of EscapeQuotes()
        command = self._sql[(Commands.CONTAINS,
                             subject != '',
                             predicate != '',
                             object != '',
                             statementUri != '',
                             sourceUri != '',
                             )]
                               
        result = command.query(self._db,
                               subject=subject,
                               predicate=predicate,
                               object=object,
                               statementUri=statementUri,
                               sourceUri=sourceUri,
                               subjectOp=self._cmp[flags.get('subjectFlags')],
                               predicateOp=self._cmp[flags.get('predicateFlags')],
                               objectOp=self._cmp[flags.get('objectFlags')],
                               statementUriOp=self._cmp[flags.get('statementUriFlags')],
                               sourceUriOp=self._cmp[flags.get('sourceUriFlags')],
                               )

        return result and result[0][0] > 0 or 0

    def bind(self, object, name, sourceUri, readAcl, deleteAcl, aclIds, aclNames):
        if not self._db:
            raise DataBaseExceptions.NoTransaction
        
        readId = readAcl and self.verifyAcl(readAcl, aclIds, aclNames)
        deleteId = deleteAcl and self.verifyAcl(deleteAcl, aclIds, aclNames)
        self._sql[Commands.BIND].execute(self._db,
                                         readAcl=readId,
                                         deleteAcl=deleteId,
                                         object=EscapeQuotes(cPickle.dumps(object)),
                                         name=EscapeQuotes(name),
                                         sourceUri=EscapeQuotes(sourceUri),
                                         )
        return

    def unbind(self, name, sourceUri, aclIds):
        if not self._db:
            raise DataBaseExceptions.NoTransaction

        self._sql[Commands.UNBIND].execute(self._db,
                                           deleteAcl=string.join(aclIds, ','),
                                           name=EscapeQuotes(name),
                                           sourceUri=EscapeQuotes(sourceUri),
                                           )
        return

    def lookup(self, name, sourceUri, aclIds):
        if not self._db:
            raise DataBaseExceptions.NoTransaction

        result = self._sql[Commands.LOOKUP].query(self._db,
                                                  readAcl=string.join(aclIds, ','),
                                                  name=EscapeQuotes(name),
                                                  sourceUri=EscapeQuotes(sourceUri),
                                                  )
        return result and cPickle.loads(result[0][0]) or None

    def keys(self, sourceUri):
        result = self._sql[Commands.KEYS].query(self._db,
                                                sourceUri=EscapeQuotes(sourceUri),
                                                )
        return map(lambda x: x[0], result or [])

    def has_key(self, name, sourceUri):
        result = self._sql[Commands.HAS_KEY].query(self._db,
                                                   name=EscapeQuotes(name),
                                                   sourceUri=EscapeQuotes(sourceUri),
                                                   )
        return result and int(result[0][0]) or 0
